/******************************************************************************/
/* Copyright (C) 2006 Jagoon Software. All rights reserved.                   */
/*                                                                            */
/* Redistribution and use in source and binary forms, with or without         */
/* modification, are permitted provided that the following conditions         */
/* are met:                                                                   */
/*                                                                            */
/*   Redistributions of source code must retain the above copyright notice,   */
/*   this list of conditions and the following disclaimer.                    */
/*                                                                            */
/*   Redistributions in binary form must reproduce the above copyright        */
/*   notice, this list of conditions and the following disclaimer in the      */
/*   documentation and/or other materials provided with the distribution.     */
/*                                                                            */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS        */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED  */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR           */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,      */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,        */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,            */
/* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF  */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING       */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS         */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.               */
/******************************************************************************/

#include "emapconv.h"

#include <d3d9.h>
#include <d3dx9.h>
#include <math.h>
#include <tchar.h>

namespace dempedit {

// ǂݍݎ̐ݒۑ\
struct LOADOPTION {
	LPCSTR pMemory;
	DWORD SizeOfMemory;
	DWORD MeshOptions;
	LPDIRECT3DDEVICE9 pDevice;
	LPD3DXALLOCATEHIERARCHY pAlloc;
	LPD3DXLOADUSERDATA pUserDataLoader;
	LPD3DXFRAME * ppFrameHeirarchy;
	LPD3DXANIMATIONCONTROLLER * ppAnimController;
};

// nʂ̃f[^iP}X̃eNX`ʒuȂǁj
#pragma pack(push, 1)
struct GROUNDTEXT {
	DWORD attrib;
	float u[4];
	float v[4];
};
#pragma pack(pop)

// nʂ̃f[^iP}X̃f[^j
#pragma pack(push, 1)
struct GROUNDCELL {
	float height[4];
	struct GROUNDTEXT text[3];
};
#pragma pack(pop)

// nʂ̃f[^
struct GROUNDDATA {
	DWORD dwWidth;
	DWORD dwHeight;
	float fSizeX;
	float fSizeZ;
	const GROUNDCELL *pCellData;
	const float *pHeightDataX;
	const float *pHeightDataY;
	DWORD NumMaterials;
	LPD3DXMATERIAL pMaterials;
};

// nʗpbV̒_FVF錾
#define GROUNDFVFDECL (D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1)

// nʗpbV̒_FVF\
struct GROUNDFVF {
	float x, y, z; // _W
	float nx, ny, nz; // @
	float u, v; // eNX`W
};

// [J֐錾
static LPD3DXFRAME LoadMesh( LOADOPTION* pOption );
static LPD3DXFRAME LoadField( LOADOPTION* pOption );
static LPD3DXFRAME LoadGround( LOADOPTION* pOption );
static LPD3DXFRAME LoadSpline( LOADOPTION* pOption );
static LPD3DXMESHCONTAINER ConvertGroundToMesh(LOADOPTION* pOption, GROUNDDATA *pGround, int x0, int y0, int x1, int y1);

// -------------------------------------
// IuWFNg̓ǂݍ݁iC֐j
// -------------------------------------

HRESULT LoadEmapFromMemory(
	LPCVOID pMemory, DWORD SizeOfMemory, DWORD MeshOptions, LPDIRECT3DDEVICE9 pDevice,
	LPD3DXALLOCATEHIERARCHY pAlloc, LPD3DXLOADUSERDATA pUserDataLoader,
	LPD3DXFRAME * ppFrameHeirarchy, LPD3DXANIMATIONCONTROLLER * ppAnimController)
{
	// ̃`FbN
	if( pMemory == NULL || pDevice == NULL || pAlloc == NULL || ppFrameHeirarchy == NULL )
		return E_INVALIDARG;

	if( SizeOfMemory < 4 ) return E_INVALIDARG;

	LOADOPTION option;
	option.pMemory          = (LPCSTR)pMemory + 4;
	option.SizeOfMemory     = SizeOfMemory - 4;
	option.MeshOptions      = MeshOptions & (~D3DXMESH_32BIT); // Ô߃NAĂ
	option.pDevice          = pDevice;
	option.pAlloc           = pAlloc;
	option.pUserDataLoader  = pUserDataLoader;
	option.ppFrameHeirarchy = ppFrameHeirarchy;
	option.ppAnimController = ppAnimController;

	LPD3DXFRAME pFrame = NULL;
	if( !memcmp( pMemory, "FLD0", 4 ) ) {
		pFrame = LoadField( &option );
	}
	if( ppFrameHeirarchy != NULL ) *ppFrameHeirarchy = pFrame;
	if( ppAnimController != NULL ) *ppAnimController = NULL;

	return (pFrame ? S_OK : E_FAIL);
}

// -------------------------------------
// IuWFNg̓ǂݍ݁iC֐j
// -------------------------------------

HRESULT LoadEmapFromFile(
	LPCTSTR Filename, DWORD MeshOptions, LPDIRECT3DDEVICE9 pDevice,
	LPD3DXALLOCATEHIERARCHY pAlloc, LPD3DXLOADUSERDATA pUserDataLoader,
	LPD3DXFRAME * ppFrameHeirarchy, LPD3DXANIMATIONCONTROLLER * ppAnimController)
{
	// ̃`FbN
	if( Filename == NULL || pDevice == NULL || pAlloc == NULL || ppFrameHeirarchy == NULL )
		return E_INVALIDARG;

	HANDLE hFile = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL,
								OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );

	if( hFile == INVALID_HANDLE_VALUE ) return E_FAIL;

	DWORD dwSize    = GetFileSize( hFile, NULL );
	HANDLE hMapping = CreateFileMapping( hFile, NULL, PAGE_READONLY, 0, 0, NULL );

	if( hMapping == NULL ) {
		CloseHandle( hFile );
		return E_FAIL;
	}

	// JgfBNg̈ړ
	TCHAR *pFile   = _tcsdup( Filename );
	DWORD  nLen    = GetCurrentDirectory( 0, NULL );
	TCHAR *pOldDir = (TCHAR*)malloc( nLen * sizeof(TCHAR) );
	TCHAR *pPath   = NULL;
	GetCurrentDirectory( nLen, pOldDir );

#ifdef _UNICODE
	pPath = wcsrchr( pFile, '\\' );
#else
	for( int i = 0; pFile[i]; i++ ) {
		if( isleadbyte( (unsigned char)pFile[i] ) && pFile[i+1] != '\0' ) {
			i++;  // Pǂݔ΂
		} else if( pFile[i] == '\\' ) {
			pPath = pFile + i;
		}
	}
#endif

	if( pPath ) {
		pPath[0] = '\0';
		SetCurrentDirectory( pFile );
	}

	// }bsÕAhX擾ĕϊs
	void *pMemory = MapViewOfFile( hMapping, FILE_MAP_READ, 0, 0, 0 );
	HRESULT ret   = E_FAIL;
	if( pMemory ) {
		ret = LoadEmapFromMemory(
			pMemory, dwSize, MeshOptions, pDevice, pAlloc,
			pUserDataLoader, ppFrameHeirarchy, ppAnimController
		);
		UnmapViewOfFile( pMemory );
	}
	CloseHandle( hMapping );
	CloseHandle( hFile );

	// JgfBNgɖ߂
	SetCurrentDirectory( pOldDir );
	free( pFile );
	free( pOldDir );

	return ret;
}

// -------------------------------------
// ̓ǂݍ(MBCS)
// -------------------------------------

static LPSTR ReadStringA( const char **ppData, DWORD *pSize ) {
	wchar_t *pData = (wchar_t*)*ppData;
	if( *pSize < 4 ) return NULL;

	DWORD nlen = *(DWORD*)pData, wlen;
	pData  += 2;
	*pSize -= 4;

	if( nlen > 1024        ) return NULL;
	if( *pSize < nlen * 2  ) return NULL;
	if( pData[nlen-1] != 0 ) return NULL;

	wlen = WideCharToMultiByte(CP_ACP, 0, pData, -1, NULL, 0, NULL, NULL);
	char *pBuf = (char*)malloc( wlen );
	WideCharToMultiByte(CP_ACP, 0, pData, -1, pBuf, wlen, NULL, NULL);

	pData  += nlen;
	*pSize -= nlen * 2;

	*ppData = (char*)pData;

	return pBuf;
}

// -------------------------------------
// ̓ǂݍ(Unicode)
// -------------------------------------

static LPWSTR ReadStringW( const char **ppData, DWORD *pSize ) {
	wchar_t *pData = (wchar_t*)*ppData;
	if( *pSize < 4 ) return NULL;

	DWORD nlen = *(DWORD*)pData;
	pData  += 2;
	*pSize -= 4;

	if( nlen > 1024        ) return NULL;
	if( *pSize < nlen * 2  ) return NULL;
	if( pData[nlen-1] != 0 ) return NULL;

	wchar_t *pBuf = _wcsdup( pData );

	pData  += nlen;
	*pSize -= nlen * 2;

	*ppData = (char*)pData;

	return pBuf;
}

// -------------------------------------
// ̓ǂݍ(Unicode / MBCSʂ̃}N)
// -------------------------------------

#ifdef _UNICODE
#define ReadString ReadStringW
#else
#define ReadString ReadStringA
#endif

// -------------------------------------
// t[̓ǂݍ
// -------------------------------------

static LPD3DXFRAME LoadObjectFrame( LOADOPTION* pOption ) {
	LPCSTR data = pOption->pMemory;
	DWORD  size = pOption->SizeOfMemory;

	if( size < 48 ) return NULL;
	if( memcmp( data, "OBJ1", 4 ) ) return NULL;

	float *scale   = (float*)(data + 4);
	float *rotate  = (float*)(data + 4 + 12);
	float *trans   = (float*)(data + 4 + 24);
	BOOL  *visible = (BOOL *)(data + 4 + 36);
	char  *field   = (char *)(data + 4 + 36 + 4);
	if( memcmp( field, "NAME", 4 ) ) return NULL;

	data += 48;
	size -= 48;

	LPSTR pName = (LPSTR)ReadStringA( &data, &size );
	if( pName == NULL ) return NULL;

	// OɔpXy[XsIh܂܂ĂꍇA
	// ۑX t@CǂݍނƎŝŕ␳
	pName = _strdup( pName );
	for( DWORD i = 0; pName[i]; i++ ) {
		if( pName[i] == ' ' || pName[i] == '.' ) pName[i] = '_';
	}

	LPD3DXFRAME pFrame = NULL;
	if( FAILED( pOption->pAlloc->CreateFrame( pName, &pFrame ) ) ) {
		// ǂݍ݂Ɏs
		free( pName );
		return NULL;
	}
	free( pName );

	// ÔNULLNAĂ
	pFrame->pMeshContainer   = NULL;
	pFrame->pFrameSibling    = NULL;
	pFrame->pFrameFirstChild = NULL;

	D3DXMATRIX mat1, mat2;
	D3DXMatrixScaling( &mat1, scale[0], scale[1], scale[2] );
	D3DXMatrixRotationYawPitchRoll(
		&mat2,
		D3DXToRadian( rotate[1] ), // y
		D3DXToRadian( rotate[0] ), // x
		D3DXToRadian( rotate[2] )  // z
	);
	D3DXMatrixMultiply( &pFrame->TransformationMatrix, &mat1, &mat2 );
	pFrame->TransformationMatrix._41 += trans[0];
	pFrame->TransformationMatrix._42 += trans[1];
	pFrame->TransformationMatrix._43 += trans[2];

	pOption->pMemory      = data;
	pOption->SizeOfMemory = size;

	return pFrame;
}

// -------------------------------------
// tB[h̓ǂݍ
// -------------------------------------

static LPD3DXFRAME LoadField( LOADOPTION* pOption ) {
	LPD3DXFRAME pFrame  = LoadObjectFrame( pOption );
	GROUNDDATA *pGround = NULL;

	if( pFrame == NULL || pOption->SizeOfMemory < 8 || memcmp( pOption->pMemory, "OBJS", 4 ) ) {
		goto err;
	}

	DWORD nchild = *(DWORD*)(pOption->pMemory + 4);

	pOption->pMemory      += 8;
	pOption->SizeOfMemory -= 8;

	for( DWORD i = 0; i < nchild; i++ ) {
		if( pOption->SizeOfMemory < 4 ) goto err;
		const char *type = pOption->pMemory;
		LPD3DXFRAME pChild;
		pOption->pMemory      += 4;
		pOption->SizeOfMemory -= 4;
		if( !memcmp( type, "FLD0", 4 ) ) {
			pChild = LoadField( pOption );
		} else if( !memcmp( type, "MAP0", 4 ) ) {
			pChild = LoadGround( pOption );
		} else if( !memcmp( type, "MSH1", 4 ) ) {
			pChild = LoadMesh( pOption );
		} else if( !memcmp( type, "SPL0", 4 ) ) {
			pChild = LoadSpline( pOption );
		} else {
			pChild = NULL;
		}
		if( pChild == NULL ) goto err;

		D3DXFrameAppendChild( pFrame, pChild );
	}

	return pFrame;

err:
	D3DXFrameDestroy( pFrame, pOption->pAlloc );
	return NULL;
}

// -------------------------------------
// bV̓ǂݍ
// -------------------------------------

static LPD3DXFRAME LoadMesh( LOADOPTION* pOption ) {
	TCHAR      *pXFile  = NULL;
	LPD3DXFRAME pFrame  = LoadObjectFrame( pOption );

	if( pFrame == NULL ) {
		return NULL;
	}

	// t@C̎擾
	if( pOption->SizeOfMemory < 4 || memcmp( pOption->pMemory, "FILE", 4 ) ) {
		D3DXFrameDestroy( pFrame, pOption->pAlloc );
		return NULL;
	}

	pOption->pMemory      += 4;
	pOption->SizeOfMemory -= 4;
	pXFile = ReadString( &pOption->pMemory, &pOption->SizeOfMemory );
	if( pXFile == NULL ) {
		D3DXFrameDestroy( pFrame, pOption->pAlloc );
		return NULL;
	}

	// bV̓ǂݍ݁iLbV͊eł肢܂j
	LPD3DXFRAME pChildFrame;
	LPD3DXANIMATIONCONTROLLER pAnimController;

	int hr = D3DXLoadMeshHierarchyFromX(
		pXFile,
		pOption->MeshOptions,
		pOption->pDevice,
		pOption->pAlloc,
		pOption->pUserDataLoader,
		&pChildFrame,
		&pAnimController
	);
	free( pXFile );

	if( FAILED( hr ) ) {
		D3DXFrameDestroy( pFrame, pOption->pAlloc );
		return NULL;
	}

	D3DXFrameAppendChild( pFrame, pChildFrame );
	if( pAnimController ) pAnimController->Release(); // T|[gÖ׉

	return pFrame;
}

// -------------------------------------
// nʂ̓ǂݍ
// -------------------------------------

static LPD3DXFRAME LoadGround( LOADOPTION* pOption ) {
	LPD3DXFRAME pFrame  = LoadObjectFrame( pOption );
	GROUNDDATA *pGround = NULL;
	LPCSTR data = pOption->pMemory;
	DWORD  size = pOption->SizeOfMemory;

	if( pFrame == NULL || size < 24 || memcmp( data, "HEAD", 4 ) ) {
		goto err;
	}

	pGround            = (GROUNDDATA*)calloc( 1, sizeof( GROUNDDATA ) );
	pGround->dwWidth   = *(DWORD*)( data +  4 );
	pGround->dwHeight  = *(DWORD*)( data +  8 );
	pGround->fSizeX    = *(float*)( data + 12 ); 
	pGround->fSizeZ    = *(float*)( data + 16 ); 
	pGround->pCellData = (const GROUNDCELL*)( data + 24 );
	if( memcmp( data + 20, "DATA", 4 ) ) {
		goto err;
	}
	if( pGround->dwWidth > 4096 || pGround->dwHeight > 4096 ) {
		// Ŝ߁ATCY4Kȏ̂̂̓G[ɂ
		goto err;
	}

	int offset = 24;

	offset += sizeof( GROUNDCELL ) * pGround->dwWidth * pGround->dwHeight;
	pGround->pHeightDataY = (const float*)( data + offset );
	offset += sizeof( float ) * 4 * pGround->dwWidth;
	pGround->pHeightDataX = (const float*)( data + offset );
	offset += sizeof( float ) * 4 * pGround->dwHeight;

	size -= offset;
	data += offset;
	if( size < 8 || memcmp( data, "TEXT", 4 ) ) {
		goto err;
	}
	pGround->NumMaterials = *(DWORD*)(data + 4);
	pGround->pMaterials   = (LPD3DXMATERIAL)calloc( pGround->NumMaterials, sizeof( D3DXMATERIAL ) );
	data += 8;
	size -= 8;
	for( DWORD i = 0; i < pGround->NumMaterials; i++ ) {
		LPSTR pTexture = ReadStringA( &data, &size );
		if( pTexture == NULL ) goto err;
		D3DXMATERIAL *pMaterial = &pGround->pMaterials[i];
		pMaterial->MatD3D.Diffuse.r = 1.0f;
		pMaterial->MatD3D.Diffuse.g = 1.0f;
		pMaterial->MatD3D.Diffuse.b = 1.0f;
		pMaterial->MatD3D.Diffuse.a = 1.0f;
		pMaterial->pTextureFilename = pTexture;
	}

	// nʂ̕ϊizj
	pFrame->pMeshContainer = ConvertGroundToMesh( pOption, pGround, 0, 0, pGround->dwWidth, pGround->dwHeight );
	if( pGround ) {
		for( DWORD i = 0; i < pGround->NumMaterials; i++ ) {
			free( pGround->pMaterials[i].pTextureFilename );
		}
		free( pGround->pMaterials );
		free( pGround );
	}

	// cf[^̈p
	pOption->SizeOfMemory  = size;
	pOption->pMemory       = data;

	return pFrame;

err:
	D3DXFrameDestroy( pFrame, pOption->pAlloc );
	if( pGround ) {
		for( DWORD i = 0; i < pGround->NumMaterials; i++ ) {
			free( pGround->pMaterials[i].pTextureFilename );
		}
		free( pGround->pMaterials );
		free( pGround );
	}
	return NULL;
}

// -------------------------------------
// ̎擾
// -------------------------------------

static void GetHeightValue( GROUNDDATA *pGround, int x, int y, int t, float *h1 ) {
	const float *h2, *h3;
	switch( t ) {
	case 0:
		h2 = pGround->pCellData[ y * pGround->dwWidth + x ].height;
		h1[0] = h2[0];
		h1[1] = h2[1];
		h1[2] = h2[2];
		h1[3] = h2[3];
		break;
	case 1:
		if( x == pGround->dwWidth - 1 ) {
			h2 = &pGround->pHeightDataY[ y * 4 ];
			h1[0] = h2[0];
			h1[1] = h2[1];
			h1[2] = h2[2];
			h1[3] = h2[3];
		} else {
			h2 = pGround->pCellData[ y * pGround->dwWidth + x    ].height;
			h3 = pGround->pCellData[ y * pGround->dwWidth + x + 1].height;
			h1[0] = h2[1];
			h1[1] = h3[0];
			h1[2] = h2[3];
			h1[3] = h3[2];
		}
		break;
	case 2:
		if( y == pGround->dwHeight - 1 ) {
			h2 = &pGround->pHeightDataX[ x * 4 ];
			h1[0] = h2[0];
			h1[1] = h2[1];
			h1[2] = h2[2];
			h1[3] = h2[3];
		} else {
			h2 = pGround->pCellData[ y     * pGround->dwWidth + x ].height;
			h3 = pGround->pCellData[ (y+1) * pGround->dwWidth + x ].height;
			h1[0] = h2[2];
			h1[1] = h2[3];
			h1[2] = h3[0];
			h1[3] = h3[1];
		}
		break;
	}
}

// -------------------------------------
// eNX`ԍ̎擾
// -------------------------------------

static int GetTextureNumber( GROUNDDATA *pGround, int x, int y, int t ) {
	return pGround->pCellData[ y * pGround->dwWidth + x ].text[t].attrib;
}

// -------------------------------------
// eNX`ʒu̎擾
// -------------------------------------

static void GetTexturePos( GROUNDDATA *pGround, int x, int y, int t, float *u, float *v ) {
	const GROUNDTEXT *pText = &pGround->pCellData[ y * pGround->dwWidth + x ].text[t];
	u[0] = pText->u[0];
	u[1] = pText->u[1];
	u[2] = pText->u[2];
	u[3] = pText->u[3];

	v[0] = pText->v[0];
	v[1] = pText->v[1];
	v[2] = pText->v[2];
	v[3] = pText->v[3];
}

// -------------------------------------
// \ǂ̔iX j
// @A0 A1 | B0 B1
// @A2 A3 | B2 B3
// @́AƂƁAɑΏۂɂȂĂ邱
// -------------------------------------

#define CanWeldSubX(v1, v2) (                            \
	fabs( v1[1] - v2[0] ) < 0.01f &&                     \
	fabs( v1[3] - v2[2] ) < 0.01f &&                     \
	fabs( (v1[0] - v1[1]) - (v2[0] - v2[1]) ) < 0.01f && \
	fabs( (v1[2] - v1[3]) - (v2[2] - v2[3]) ) < 0.01f)

// -------------------------------------
// \ǂ̔iY j
// @0 1
// @2 3
// @----
// @0 1
// @2 3
// @́AƂƁAɑΏۂɂȂĂ邱
// -------------------------------------

#define CanWeldSubY(v1, v2) (                            \
	fabs( v1[2] - v2[0] ) < 0.01f &&                     \
	fabs( v1[3] - v2[1] ) < 0.01f &&                     \
	fabs( (v1[0] - v1[2]) - (v2[0] - v2[2]) ) < 0.01f && \
	fabs( (v1[1] - v1[3]) - (v2[1] - v2[3]) ) < 0.01f)

// -------------------------------------
// \ǂ̔iX j
// -------------------------------------

static bool CanWeldX( GROUNDDATA *pGround, int x, int y, int t ) {
	float h1[4], h2[4];
	int a1, a2;

	// ̘AmF
	GetHeightValue( pGround, x  , y, t, h1 );
	GetHeightValue( pGround, x+1, y, t, h2 );
	if( !CanWeldSubX( h1, h2 ) ) return false;

	// eNX`̘AmF
	a1 = GetTextureNumber( pGround, x  , y, t );
	a2 = GetTextureNumber( pGround, x+1, y, t );
	if( a1 != a2 ) {
		// eNX`̕sv
		return false;
	} else if( a1 ) {
		// eNX`̘AmF
		float u1[4], u2[4], v1[4], v2[4];
		GetTexturePos( pGround, x  , y, t, u1, v1 );
		GetTexturePos( pGround, x+1, y, t, u2, v2 );
		if( !CanWeldSubX( u1, u2 ) ) return false;
		if( !CanWeldSubX( v1, v2 ) ) return false;
	}
	return true;
}

// -------------------------------------
// \ǂ̔iY j
// -------------------------------------

static bool CanWeldY( GROUNDDATA *pGround, int x, int y, int t ) {
	float h1[4], h2[4];
	int a1, a2;

	// ̘AmF
	GetHeightValue( pGround, x, y  , t, h1 );
	GetHeightValue( pGround, x, y+1, t, h2 );
	if( !CanWeldSubY( h1, h2 ) ) return false;

	// eNX`̘AmF
	a1 = GetTextureNumber( pGround, x, y  , t );
	a2 = GetTextureNumber( pGround, x, y+1, t );
	if( a1 != a2 ) {
		// eNX`̕sv
		return false;
	} else if( a1 ) {
		// eNX`̘AmF
		float u1[4], u2[4], v1[4], v2[4];
		GetTexturePos( pGround, x, y  , t, u1, v1 );
		GetTexturePos( pGround, x, y+1, t, u2, v2 );
		if( !CanWeldSubY( u1, u2 ) ) return false;
		if( !CanWeldSubY( v1, v2 ) ) return false;
	}
	return true;
}

// -------------------------------------
// bVւ̒_̒ǉiwp[p}Nj
// -------------------------------------

#define AddVertex( x, z, y, u, v ) AddVertexSub( pGround, &pVertex, pBaseVertex, sx, pPos, x, z, y, u, v )

// -------------------------------------
// bVւ̒_̒ǉ
// -------------------------------------

static int AddVertexSub(
	GROUNDDATA *pGround, GROUNDFVF **ppVertex, GROUNDFVF *pBaseVertex, DWORD sx,
	DWORD *pPos, int x, int z, float y, float u, float v )
{
	// _ɂ钸_̈ꗗiőPQj
	DWORD *pIndex = &pPos[ 12 * ( (z) * (sx+1) + (x) ) ];

	while(1) {
		int j = *pIndex;
		if( j == 0 ) {
			// gpf[^̐擪Ȃ̂ŁA͒߂邱Ƃɂ
			j = (int)(*ppVertex - pBaseVertex);
			pBaseVertex[ j ].x = (float)x * pGround->fSizeX;
			pBaseVertex[ j ].y = y;
			pBaseVertex[ j ].z = (float)z * pGround->fSizeZ;
			pBaseVertex[ j ].u = u;
			pBaseVertex[ j ].v = v;
			(*ppVertex)++;
			*pIndex = j + 1;
			return j;
		}

		if( fabs( pBaseVertex[ j - 1 ].y - y ) < 0.01f &&
			fabs( pBaseVertex[ j - 1 ].u - u ) < 0.01f &&
			fabs( pBaseVertex[ j - 1 ].v - v ) < 0.01f)
		{
			// \Ȓ_̂ŁAԂ
			return j - 1;
		}
		pIndex++;
	}

}

// -------------------------------------
// bVւ̕ϊi͈͎wEœKLj
// @O[fB[ȃASYpĂ̂ŁA
// @ȂɑœKȌʂ͓ȂȂcB
// -------------------------------------

static LPD3DXMESHCONTAINER ConvertGroundToMesh(LOADOPTION* pOption, GROUNDDATA *pGround, int x0, int y0, int x1, int y1) {
	int sx = (x1 - x0);
	int sy = (y1 - y0);
	GROUNDFVF *pVertex  = (GROUNDFVF*)malloc( 12 * sx * sy * sizeof(GROUNDFVF) );
	DWORD      *pIndex  = (DWORD    *)malloc( 18 * sx * sy * sizeof(DWORD)     );
	DWORD      *pAttrib = (DWORD    *)malloc(  6 * sx * sy * sizeof(DWORD)     );
	BYTE       *pFlag   = (BYTE     *)calloc(  3 * sx * sy,  sizeof(BYTE)      );     // _o͂ǂ
	DWORD      *pPos    = (DWORD    *)calloc( 12 * (sx+1) * (sy+1), sizeof(DWORD) ); // xzW̓_ꗗ
	GROUNDFVF *pBaseVertex  = pVertex;
	DWORD      *pBaseIndex  = pIndex;
	DWORD      *pBaseAttrib = pAttrib;

	for(DWORD i = 1; i <= pGround->NumMaterials; i++) {	// obt@ƂɃ\[g
		for(int y = y0; y < y1; y++) {
			for(int x = x0; x < x1; x++) {
				int attrib;
				float height[4];

				// t == 0 ()
				attrib = GetTextureNumber( pGround, x, y, 0 );

				// obt@vAɏo͍ς݂ł͂Ȃꍇ͏o
				if( i == attrib && !pFlag[ 3 * (y - y0) * sx + 3 * (x - x0) ] ) {

					// o͈͂͂̌
					int maxw = 1, maxh = 1;

					// ő̕߂
					for( ; maxw + x < x1; maxw++) {
						if( pFlag[ 3 * (y - y0) * sx + 3 * (x - x0 + maxw) ] ) break;
						if( !CanWeldX( pGround, x + maxw - 1, y, 0 ) ) break;
					}

					// A\Ȕ͈͂̒ŁA( * ) őɂȂ̂I肷
					int limit = maxw; // ݂̍ő啝
					for(int curh = 0; curh + y < y1; curh++) {
						int curw;
						for(curw = 0; curw + x < x1 && curw < limit; curw++) {
							if( pFlag[ 3 * (y - y0 + curh) * sx + 3 * (x - x0 + curw) ] ) break;
							if( curw > 0 && !CanWeldX(pGround, x + curw - 1, y + curh, 0)) break;
							if( curh > 0 && !CanWeldY(pGround, x + curw, y + curh - 1, 0)) break;
						}
						if( curw == 0 ) {
							// ȏcɘAłȂ̂őł؂
							break;
						} else if( curw * (curh+1) > maxw * maxh ) {
							maxw = curw;
							maxh = curh + 1;
						}
						limit = curw; // ő啝XV
					}

					// o͍ς݃tOt
					for(int curw = 0; curw < maxw; curw++) {
						for(int curh = 0; curh < maxh; curh++) {
							pFlag[ 3 * (y - y0 + curh) * sx + 3 * (x - x0 + curw) ] = 1;
						}
					}

					float h[4][4], u[4][4], v[4][4];
					int idx[4];
					GetHeightValue( pGround, x           , y, 0, h[0] );
					GetHeightValue( pGround, x + maxw - 1, y, 0, h[1] );
					GetHeightValue( pGround, x           , y + maxh - 1, 0, h[2] );
					GetHeightValue( pGround, x + maxw - 1, y + maxh - 1, 0, h[3] );

					GetTexturePos( pGround, x           , y, 0, u[0], v[0] );
					GetTexturePos( pGround, x + maxw - 1, y, 0, u[1], v[1] );
					GetTexturePos( pGround, x           , y + maxh - 1, 0, u[2], v[2] );
					GetTexturePos( pGround, x + maxw - 1, y + maxh - 1, 0, u[3], v[3] );

					// _f[^
					idx[0] = AddVertex( x - x0       , y - y0       , h[0][0], u[0][0], v[0][0] );
					idx[1] = AddVertex( x - x0 + maxw, y - y0       , h[1][1], u[1][1], v[1][1] );
					idx[2] = AddVertex( x - x0       , y - y0 + maxh, h[2][2], u[2][2], v[2][2] );
					idx[3] = AddVertex( x - x0 + maxw, y - y0 + maxh, h[3][3], u[3][3], v[3][3] );

					// CfbNXf[^
					pIndex[0] = idx[1];
					pIndex[1] = idx[0];
					pIndex[2] = idx[2];
					pIndex[3] = idx[1];
					pIndex[4] = idx[2];
					pIndex[5] = idx[3];

					// ߃eNX`o͖
					pAttrib[0] = i - 1;
					pAttrib[1] = i - 1;

					pIndex  += 6;
					pAttrib += 2;
				}

				// t == 1 (E)

				// obt@vAɏo͍ς݂ł͂Ȃꍇ͏o
				attrib = GetTextureNumber( pGround, x, y, 1 );
				GetHeightValue( pGround, x, y, 1, height );
				if( i == attrib && !pFlag[ 3 * (y - y0) * sx + 3 * (x - x0) + 1 ] &&
					!( fabs( height[0] - height[1] ) < 0.01f && fabs( height[2] - height[3] ) < 0.01f ) )
				{

					// o͈͂͂̌
					int maxh = 1;

					// ő̍߂
					for( ; maxh + y < y1; maxh++) {
						if( pFlag[ 3 * (y - y0 + maxh) * sx + 3 * (x - x0) + 1 ] ) break;
						if( !CanWeldY( pGround, x, y + maxh - 1, 1 ) ) break;
					}

					// o͍ς݃tOt
					for(int curh = 0; curh < maxh; curh++) {
						pFlag[ 3 * (y - y0 + curh) * sx + 3 * (x - x0) + 1 ] = 1;
					}

					float h[2][4], u[2][4], v[2][4], u2, v2;
					int face;
					int out[4];
					int idx[4];

					GetHeightValue( pGround, x, y, 1, h[0] );
					GetHeightValue( pGround, x, y + maxh - 1, 1, h[1] );

					GetTexturePos( pGround, x, y, 1, u[0], v[0] );
					GetTexturePos( pGround, x, y + maxh - 1, 1, u[1], v[1] );

					if( fabs( h[0][0] - h[0][1] ) < 0.01f ) {
						// Op` 1, 3, 2
						out[0] = 0; out[1] = 1; out[2] = 3; out[3] = 2; face = 1;
					} else if( fabs( h[1][2] - h[1][3] ) < 0.01f ) {
						// Op` 0, 1, 2
						out[0] = 1; out[1] = 2; out[2] = 3; out[3] = 0; face = 1;
					} else {
						// lp`
						out[0] = 1; out[1] = 2; out[2] = 3; out[3] = 4; face = 2;
					}

					// _f[^ij
					if( out[0] ) {
						u2 = ( out[1] ? u[0][0] : (u[0][0] + u[0][1]) / 2.0f );
						v2 = ( out[1] ? v[0][0] : (v[0][0] + v[0][1]) / 2.0f );
						idx[0] = AddVertex( x - x0 + 1, y - y0, h[0][0], u2, v2 );
					}

					// _f[^iEj
					if( out[1] ) {
						u2 = ( out[0] ? u[0][1] : (u[0][0] + u[0][1]) / 2.0f );
						v2 = ( out[0] ? v[0][1] : (v[0][0] + v[0][1]) / 2.0f );
						idx[1] = AddVertex( x - x0 + 1, y - y0, h[0][1], u2, v2 );
					}

					// _f[^ij
					if( out[2] ) {
						u2 = ( out[3] ? u[1][2] : (u[1][2] + u[1][3]) / 2.0f );
						v2 = ( out[3] ? v[1][2] : (v[1][2] + v[1][3]) / 2.0f );
						idx[2] = AddVertex( x - x0 + 1, y - y0 + maxh, h[1][2], u2, v2 ); 
					}

					// _f[^iEj
					if( out[3] ) {
						u2 = ( out[2] ? u[1][3] : (u[1][2] + u[1][3]) / 2.0f );
						v2 = ( out[2] ? v[1][3] : (v[1][2] + v[1][3]) / 2.0f );
						idx[3] = AddVertex( x - x0 + 1, y - y0 + maxh, h[1][3], u2, v2 );
					}

					// _f[^ɃCfbNXӂ
					for(int j = 0; j < 4; j++) {
						if( out[j] == 1+1 ) pIndex[ 0 ] = idx[ j ];
						if( out[j] == 0+1 ) pIndex[ 1 ] = idx[ j ];
						if( out[j] == 2+1 ) pIndex[ 2 ] = idx[ j ];
					}
					pAttrib[0] = i - 1;
					if( face == 1 ) {
						pIndex  += 3;
						pAttrib += 1;
					} else {
						// _f[^ɃCfbNXӂ
						for(int j = 0; j < 4; j++) {
							if( out[j] == 1+1 ) pIndex[ 3 ] = idx[ j ];
							if( out[j] == 2+1 ) pIndex[ 4 ] = idx[ j ];
							if( out[j] == 3+1 ) pIndex[ 5 ] = idx[ j ];
						}
						pAttrib[1] = i - 1;
						pIndex  += 6;
						pAttrib += 2;
					}
				}

				// t == 2 (  )

				// obt@vAɏo͍ς݂ł͂Ȃꍇ͏o
				attrib = GetTextureNumber( pGround, x, y, 2 );
				GetHeightValue( pGround, x, y, 2, height );
				if( i == attrib && !pFlag[ 3 * (y - y0) * sx + 3 * (x - x0) + 2 ] &&
					!( fabs( height[0] - height[2] ) < 0.01f && fabs( height[1] - height[3] ) < 0.01f ) )
				{
					// o͈͂͂̌
					int maxw = 1;

					// ő̕߂
					for( ; maxw + x < x1; maxw++) {
						if( pFlag[ 3 * (y - y0) * sx + 3 * (x - x0 + maxw) + 2 ] ) break;
						if( !CanWeldX( pGround, x + maxw - 1, y, 2 ) ) break;
					}

					// o͍ς݃tOt
					for(int curw = 0; curw < maxw; curw++) {
						pFlag[ 3 * (y - y0) * sx + 3 * (x - x0 + curw) + 2 ] = 1;
					}

					float h[2][4], u[2][4], v[2][4], u2, v2;
					int face;
					int idx[4], out[4];

					GetHeightValue( pGround, x           , y, 2, h[0] );
					GetHeightValue( pGround, x + maxw - 1, y, 2, h[1] );

					GetTexturePos( pGround, x           , y, 2, u[0], v[0] );
					GetTexturePos( pGround, x + maxw - 1, y, 2, u[1], v[1] );

					if( fabs( h[0][0] - h[0][2] ) < 0.01f ) {
						// Op` 2, 1, 3
						out[0] = 0; out[1] = 2; out[2] = 1; out[3] = 3; face = 1;
					} else if( fabs( h[1][1] - h[1][3] ) < 0.01f ) {
						// Op` 0, 1, 2
						out[0] = 1; out[1] = 2; out[2] = 3; out[3] = 0; face = 1;
					} else {
						// lp`
						out[0] = 1; out[1] = 2; out[2] = 3; out[3] = 4; face = 2;
					}

					// _f[^ij
					if( out[0] ) {
						u2 = ( out[2] ? u[0][0] : (u[0][0] + u[0][2]) / 2.0f );
						v2 = ( out[2] ? v[0][0] : (v[0][0] + v[0][2]) / 2.0f );
						idx[0] = AddVertex( x - x0, y - y0 + 1, h[0][0], u2, v2 );
					}

					// _f[^iEj
					if( out[1] ) {
						u2 = ( out[3] ? u[1][1] : (u[1][1] + u[1][3]) / 2.0f );
						v2 = ( out[3] ? v[1][1] : (v[1][1] + v[1][3]) / 2.0f );
						idx[1] = AddVertex( x - x0 + maxw, y - y0 + 1, h[1][1], u2, v2 );
					}

					// _f[^ij
					if( out[2] ) {
						u2 = ( out[0] ? u[0][2] : (u[0][0] + u[0][2]) / 2.0f );
						v2 = ( out[0] ? v[0][2] : (v[0][0] + v[0][2]) / 2.0f );
						idx[2] = AddVertex( x - x0, y - y0 + 1, h[0][2], u2, v2 );
					}

					// _f[^iEj
					if( out[3] ) {
						u2 = ( out[1] ? u[1][3] : (u[1][1] + u[1][3]) / 2.0f );
						v2 = ( out[1] ? v[1][3] : (v[1][1] + v[1][3]) / 2.0f );
						idx[3] = AddVertex( x - x0 + maxw, y - y0 + 1, h[1][3], u2, v2 );
					}

					// _f[^ɃCfbNXӂ
					for(int j = 0; j < 4; j++) {
						if( out[j] == 1+1 ) pIndex[ 0 ] = idx[ j ];
						if( out[j] == 0+1 ) pIndex[ 1 ] = idx[ j ];
						if( out[j] == 2+1 ) pIndex[ 2 ] = idx[ j ];
					}

					pAttrib[0] = i - 1;
					if( face == 1 ) {
						pIndex  += 3;
						pAttrib += 1;
					} else {
						// _f[^ɃCfbNXӂ
						for(int j = 0; j < 4; j++) {
							if( out[j] == 1+1 ) pIndex[ 3 ] = idx[ j ];
							if( out[j] == 2+1 ) pIndex[ 4 ] = idx[ j ];
							if( out[j] == 3+1 ) pIndex[ 5 ] = idx[ j ];
						}
						pAttrib[1] = i - 1;
						pIndex  += 6;
						pAttrib += 2;
					}
				}
			}
		}
	}

	int hr;
	DWORD nfaces  = (int)(pIndex  - pBaseIndex ) / 3; // _
	DWORD nvertex = (int)(pVertex - pBaseVertex);     // ʐ
	LPD3DXMESH pMesh = NULL;
	LPD3DXMESHCONTAINER pRet = NULL;

	if( nfaces == 0 ) {
		// bV쐬Ȃ
		hr = E_FAIL;
	} else if( nvertex >= 65535 ) {
		// _64K zꍇA32bit CfbNXp
		DWORD *pIdx;
		hr = D3DXCreateMeshFVF(
			nfaces, nvertex, pOption->MeshOptions | D3DXMESH_32BIT,
			GROUNDFVFDECL, pOption->pDevice, &pMesh
		);
		if( SUCCEEDED( hr ) ) {
			pMesh->LockIndexBuffer( 0, (void**)&pIdx );
			memcpy( pIdx, pBaseIndex, sizeof(DWORD) * nfaces * 3 );
			pMesh->UnlockIndexBuffer();
		}
	} else {
		// CfbNX16bit
		WORD *pIdx;
		hr = D3DXCreateMeshFVF(
			nfaces, nvertex, pOption->MeshOptions,
			GROUNDFVFDECL, pOption->pDevice, &pMesh
		);
		if( SUCCEEDED( hr ) ) {
			pMesh->LockIndexBuffer( 0, (void**)&pIdx );
			for(DWORD i = 0; i < nfaces * 3; i++) {
				pIdx[i] = (WORD)pBaseIndex[i];
			}
			pMesh->UnlockIndexBuffer();
		}
	}

	if( SUCCEEDED( hr ) ) {
		void *pBuf;
		LPD3DXMESH pNewMesh;
		// _obt@̃Rs[
		pMesh->LockVertexBuffer(0, &pBuf );
		memcpy( pBuf, pBaseVertex, sizeof(GROUNDFVF) * nvertex );
		pMesh->UnlockVertexBuffer();

		// obt@̃Rs[
		pMesh->LockAttributeBuffer(0, (DWORD**)&pBuf);
		memcpy( pBuf, pBaseAttrib, sizeof(DWORD) * nfaces );
		pMesh->UnlockAttributeBuffer();

		// אږʏ̎Zo ( pBaseIndex ėp )
		pMesh->GenerateAdjacency( 0.001f, pBaseIndex );

		// Bowtie ̏
		if( SUCCEEDED( D3DXCleanMesh( D3DXCLEAN_BOWTIES, pMesh, pBaseIndex, &pNewMesh, pBaseIndex, NULL ) ) ) {
			pMesh->Release();
			pMesh = pNewMesh;
		}

		// @̌vZ
		D3DXComputeNormals( pMesh, pBaseIndex );

		// bVRei̍쐬
		D3DXMESHDATA mesh;
		mesh.Type   = D3DXMESHTYPE_MESH;
		mesh.pMesh  = pMesh;
		hr = pOption->pAlloc->CreateMeshContainer(
			NULL, &mesh, pGround->pMaterials, NULL, pGround->NumMaterials,
			pBaseIndex, NULL, &pRet
		);
		if( FAILED( hr ) ) {
			pRet = NULL;
		}
		pMesh->Release(); // ̂̓}i[
	}

	// obt@̊J
	free( pBaseVertex );
	free( pBaseIndex );
	free( pBaseAttrib );
	free( pFlag );
	free( pPos );

	return pRet;
}

// -------------------------------------
// XvCȐ̓ǂݍ
// -------------------------------------

static LPD3DXFRAME LoadSpline( LOADOPTION* pOption ) {
	LPD3DXFRAME pFrame = LoadObjectFrame( pOption );

	if( pFrame == NULL || pOption->SizeOfMemory < 8 ) {
		D3DXFrameDestroy( pFrame, pOption->pAlloc );
		return NULL;
	}

	DWORD nPoints = *(DWORD*)(pOption->pMemory + 4);
	DWORD size    = 8 + nPoints * sizeof( D3DXVECTOR3 );

	if( memcmp( pOption->pMemory, "DATA", 4 ) ||
		nPoints > 65535 || size > pOption->SizeOfMemory )
	{
		D3DXFrameDestroy( pFrame, pOption->pAlloc );
		return NULL;
	}

	// XvCȐ̓ǂݍ݂ɂẮAe҂̕ɂC܂B
	const D3DXVECTOR3 *pPoints = (D3DXVECTOR3*)(pOption->pMemory + 8);
	for( DWORD i = 0; i < nPoints; i++ ) {
		const D3DXVECTOR3 *pVec = &pPoints[i];
		// ...
	}

	pOption->pMemory      += size;
	pOption->SizeOfMemory -= size;

	return pFrame;
}

} /* namespace dempedit */
