/******************************************************************************/
/* Copyright (C) 2006 Jagoon Software. All rights reserved.                   */
/*                                                                            */
/* Redistribution and use in source and binary forms, with or without         */
/* modification, are permitted provided that the following conditions         */
/* are met:                                                                   */
/*                                                                            */
/*   Redistributions of source code must retain the above copyright notice,   */
/*   this list of conditions and the following disclaimer.                    */
/*                                                                            */
/*   Redistributions in binary form must reproduce the above copyright        */
/*   notice, this list of conditions and the following disclaimer in the      */
/*   documentation and/or other materials provided with the distribution.     */
/*                                                                            */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS        */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED  */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR */
/* PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR           */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,      */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,        */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,            */
/* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF  */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING       */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS         */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.               */
/******************************************************************************/

#include "emapluna.h"
#include "emapconv.h"

namespace dempedit {

// -------------------------------------
// bVǂݍݗpNX
// -------------------------------------

class CMeshAllocator : public ID3DXAllocateHierarchy {
public:
	CMeshAllocator() {
	}

	~CMeshAllocator() {
	}

	HRESULT __stdcall CreateFrame(LPCSTR Name, LPD3DXFRAME * ppNewFrame) {
		LPD3DXFRAME pFrame = new D3DXFRAME;
		ZeroMemory( pFrame, sizeof(D3DXFRAME) );
		pFrame->Name = (Name == NULL ? NULL : _strdup( Name ) );

		*ppNewFrame = pFrame;
		return D3D_OK;
	}

	HRESULT __stdcall CreateMeshContainer(
		LPCSTR Name, CONST D3DXMESHDATA * pMeshData,
		CONST D3DXMATERIAL * pMaterials,
		CONST D3DXEFFECTINSTANCE * pEffectInstances,
		DWORD NumMaterials,
		CONST DWORD * pAdjacency,
		LPD3DXSKININFO pSkinInfo,
		LPD3DXMESHCONTAINER * ppNewMeshContainer)
	{
		D3DXMESHCONTAINER *pMeshContainer = new D3DXMESHCONTAINER;
		ZeroMemory( pMeshContainer, sizeof(D3DXMESHCONTAINER) );
		pMeshContainer->Name         = (Name == NULL ? NULL : _strdup( Name ));
		pMeshContainer->MeshData     = *pMeshData;
		pMeshContainer->pMaterials   = new D3DXMATERIAL[ NumMaterials ];
//	    pMeshContainer->pEffects     = pEffectInstances;
		pMeshContainer->NumMaterials = NumMaterials;
//		pMeshContainer->pAdjacency   = pAdjacency;
//		pMeshContainer->pSkinInfo    = pSkinInfo;

		memcpy( pMeshContainer->pMaterials, pMaterials, sizeof(D3DXMATERIAL) * NumMaterials );

		// }eÃRs[
		for( DWORD i = 0; i < NumMaterials; i++ ) {
			pMeshContainer->pMaterials[i].pTextureFilename = _strdup( pMaterials[i].pTextureFilename );
		}

		switch( pMeshContainer->MeshData.Type ) {
		case D3DXMESHTYPE_MESH:
			pMeshData->pMesh->AddRef();
			break;
		case D3DXMESHTYPE_PMESH:
			pMeshData->pPMesh->AddRef();
			break;
		case D3DXMESHTYPE_PATCHMESH:
			pMeshData->pPatchMesh->AddRef();
			break;
		}

		*ppNewMeshContainer = pMeshContainer;

		return D3D_OK;
	}

	HRESULT __stdcall DestroyFrame(LPD3DXFRAME pFrame) {
		free( pFrame->Name );
		delete pFrame;
		return D3D_OK;
	}

	HRESULT __stdcall DestroyMeshContainer(LPD3DXMESHCONTAINER pMeshContainer) {
		// bV̉
		D3DXMESHDATA* pMeshData = &pMeshContainer->MeshData;
		switch(pMeshData->Type ) {
		case D3DXMESHTYPE_MESH:
			pMeshData->pMesh->Release();
			break;
		case D3DXMESHTYPE_PMESH:
			pMeshData->pPMesh->Release();
			break;
		case D3DXMESHTYPE_PATCHMESH:
			pMeshData->pPatchMesh->Release();
			break;
		}

		// }eA
		LPD3DXMATERIAL pMaterial = pMeshContainer->pMaterials;
		for( DWORD i = 0; i < pMeshContainer->NumMaterials; i++ ) {
			free( pMaterial[i].pTextureFilename );
		}

		free( pMeshContainer->Name );

		delete []pMaterial;
		delete pMeshContainer;

		return D3D_OK;
	}
};

// -------------------------------------
// t@CXV̎擾
// -------------------------------------

BOOL GetFileWriteTime( LPCSTR pFileName, LPFILETIME lpLastWriteTime ) {
	BOOL   result;
	HANDLE hFile = ::CreateFile(
		pFileName ,GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL
	);
	if( hFile == INVALID_HANDLE_VALUE ) return FALSE;

	result = GetFileTime( hFile, NULL, NULL, lpLastWriteTime );

	CloseHandle( hFile );

	return result;
}

// -------------------------------------
// }bvf[^̓ǂݍ for Luna
// -------------------------------------

LFIELD LoadFieldFromFile( const char *pFileName, Sint32 TreeDepth, Sint32 ZoomSize,
						  eSurfaceFormat Format, Bool IsBumpMap, D3DCOLOR KeyColor )
{
	HRESULT hr;
	char path_emap[_MAX_PATH + 1]; // emap̃pX
	char path_x   [_MAX_PATH + 3]; // ".x"̕]vɊm
	FILETIME time_emap, time_x;    // t@C̍XV

	Luna::LogoutKeyword( "Dempedit::ConvertToLunaField" );
	Luna::Logout( " : [ %s ] : ", pFileName );

	// LunaFileoRŃt@C̃pX߂
	if( !LunaFile::GetFilePath( path_emap, pFileName ) ) {
		Luna::LogoutError( "ǂݍݎs(LunaFile::GetFilePath)" );
		Luna::LogoutNextLine();
		return INVALID_FIELD;
	}
	path_emap[_MAX_PATH] = '\0';

	// X t@Cp̃pX߂
	{
		strcpy( path_x, path_emap );
		char *p = path_x + strlen( path_x );
		// 񒆂̍Ō̃sIḧʒu߂
		for( int i = 0; path_x[i]; i++ ) {
			if( isleadbyte( (unsigned char)path_x[i] ) && path_x[i+1] != '\0' ) {
				i++; // Pǂݔ΂
			} else if( path_x[i] == '.' ) {
				p = path_x + i;
			}
		}

		strcpy( p, ".x" );
	}

	// EMAP̍XV擾
	if( !GetFileWriteTime( path_emap, &time_emap ) ) {
		// XV̎擾Ɏs
		Luna::LogoutError( "ǂݍݎs(dempedit::GetFileWriteTime)" );
		Luna::LogoutNextLine();
		return INVALID_FIELD;
	}

	// EMAP̍XVX ̍XVȍ~ȂRo[g
	if( !GetFileWriteTime( path_x, &time_x ) ||
		CompareFileTime( &time_emap, &time_x ) > 0 )
	{
		LPD3DXFRAME pFrame;
		CMeshAllocator allocator;

		// EMAPLPD3DXFRAMEɕϊ
		hr = LoadEmapFromFile( path_emap, D3DXMESH_SYSTEMMEM, Luna3D::GetDevice(),
							   &allocator, NULL, &pFrame, NULL);

		if( FAILED( hr ) ) {
			Luna::LogoutError( "ǂݍݎs(dempedit::LoadEmapFromFile)" );
			Luna::LogoutNextLine();
			return INVALID_FIELD;
		}

		// X t@CƂĈ[oiLunaFieldɐp̊֐ׁj
		hr = D3DXSaveMeshHierarchyToFile( path_x, D3DXF_FILEFORMAT_BINARY, pFrame, NULL, NULL );

		// t[̉
		D3DXFrameDestroy( pFrame, &allocator );

		if( FAILED( hr ) ) {
			Luna::LogoutError( "ǂݍݎs(D3DXSaveMeshHierarchyToFile)" );
			Luna::LogoutNextLine();
			return INVALID_FIELD;
		}
	}
	Luna::LogoutOK( "Success!" );
	Luna::LogoutNextLine();

	return LunaField::CreateFromFile( path_x, TreeDepth, ZoomSize, Format, IsBumpMap, KeyColor );
}

} /* namespace dempedit */
